import pandas as pd
import pyranges as pr
import itertools
import sys
from plotly.colors import DEFAULT_PLOTLY_COLORS as plcolors
import gzip
import logging


class Transcript(object):
    def __init__(self, transcript, gene, exon_tuples, strand):
        self.transcript = transcript
        self.gene = gene
        self.exon_tuples = list(exon_tuples)
        self.strand = strand
        self.marker = "triangle-right" if self.strand == "+" else "triangle-left"
        self.begin = min(list(itertools.chain.from_iterable(self.exon_tuples)))
        self.end = max(list(itertools.chain.from_iterable(self.exon_tuples)))
        self.color = ""


def open_gtf(gtff):
    """
    Open the gtf, using gzip if it's compressed
    based on extension
    """
    return gzip.open(gtff, 'rt') if gtff.endswith('.gz') else open(gtff)


def good_record(line, chromosome):
    '''
    Filtering on the gtf lines
    by checking for right chromosome and right feature type
    '''
    if line.startswith(chromosome) \
            and line.split('\t')[2] in ['exon', 'gene']:
        return True
    else:
        return False


def get_features(gtfline, type="gtf"):
    """
    Extract the desirable features from a gtf record
    """
    chromosome, _, _, begin, end, _, strand, _, attributes = gtfline.split('\t')
    gene, transcript = parse_attributes(attributes.rstrip(), type=type)
    return [chromosome, int(begin), int(end), strand, gene, transcript]


def parse_attributes(attributes, type="gtf"):
    """
    Parse the attributes string of gtf record
    Return the values corresponding to gene_name and transcript_id
    """
    attribute_delimiter = {'gtf': '; ', 'gff': ';'}
    kv_delimiter = {'gtf': ' ', 'gff': '='}
    info = {i.split(kv_delimiter[type])[0]: i.split(kv_delimiter[type])[1].replace('"', '')
            for i in attributes.split(attribute_delimiter[type])
            if i.startswith(('gene_name', 'transcript_id', 'locus_tag'))}
    if "gene_name" in info.keys():
        return info.get("gene_name"), info.get("transcript_id")
    else:
        return info.get("locus_tag"), info.get("locus_tag")


def transcripts_in_window(df, window, feature='transcript'):
    """
    Return the transcript names for which
    either the end or the begin of an exon is within the window
    """
    return df.loc[df['begin'].between(window.begin, window.end)
                  | df['end'].between(window.begin, window.end), feature] \
        .unique()


def assign_colors_to_genes(transcripts):
    genes = set([t.gene for t in transcripts])
    colordict = {g: c for g, c in zip(genes, plcolors * 100)}
    for t in transcripts:
        t.color = colordict[t.gene]


def parse_annotation(gtff, window, simplify=False):
    """
    Parse the gtff and select the relevant region as determined by the window
    return as Transcript objects
    """
    type = annot_file_sniffer(gtff)
    logging.info(f"Parsing {type} file...")
    df = pd.DataFrame(data=[get_features(line, type=type)
                            for line in open_gtf(gtff) if good_record(line, window.chromosome)],
                      columns=['chromosome', 'begin', 'end', 'strand', 'gene', 'transcript'])
    logging.info("Loaded GTF file, processing...")
    if simplify:
        df.drop_duplicates(subset=['chromosome', 'begin', 'end', 'gene'], inplace=True)
        res = []
        for g in transcripts_in_window(df, window, feature='gene'):
            gtable = df.loc[df["gene"] == g]
            if len(gtable):
                res.append(
                    Transcript(transcript=gtable["gene"].tolist()[0],
                               gene=gtable["gene"].tolist()[0],
                               exon_tuples=gtable.loc[:, ["begin", "end"]].sort_values("begin")
                               .itertuples(index=False,
                                           name=None),
                               strand=gtable["strand"].tolist()[0])
                )
        sys.stderr.write(f"Found {len(res)} gene(s) in the region.\n")
        logging.info(f"Found {len(res)} gene(s) in the region.\n")
    else:
        res = []
        for t in transcripts_in_window(df, window, feature="transcript"):
            tr = df.loc[df["transcript"] == t]
            res.append(
                Transcript(transcript=t,
                           gene=tr["gene"].tolist()[0],
                           exon_tuples=tr.loc[:, ["begin", "end"]].sort_values("begin")
                                                                  .itertuples(index=False,
                                                                              name=None),
                           strand=tr["strand"].tolist()[0])
            )
        sys.stderr.write(f"Found {len(res)} transcript(s) in the region.\n")
        logging.info(f"Found {len(res)} transcript(s) in the region.\n")
    assign_colors_to_genes(res)
    return res


def annot_file_sniffer(annot_file):
    """
    Figure out type of annotation file

    Right not just lazily focus on the extension
    """
    if annot_file.endswith(('.gtf', '.gtf.gz')):
        return 'gtf'
    elif annot_file.endswith(('.gff', '.gff.gz')):
        return 'gff'
    else:
        sys.exit("ERROR: unrecognized extension of the annotation file.\n"
                 "Supported are gtf, gtf.gz, gff and gff.gz")


def parse_bed(bed, window):
    logging.info("Parsing BED file")
    gr = pr.read_bed(bed)[window.chromosome, window.begin:window.end]
    df = gr.unstrand().df
    df = df.drop(columns=["Chromosome", "Score", "Strand"], errors='ignore')
    if "Name" not in df.columns:
        df["Name"] = "noname"
    return df.itertuples(index=False, name=None)



from argparse import ArgumentParser, SUPPRESS
import sys
from math import ceil
from methplotlib.version import __version__
from datetime import datetime as dt
from time import time
import logging
import plotly
from pathlib import Path
from itertools import chain


class Region(object):
    def __init__(self, region, fasta=None):
        if ':' in region:
            try:
                self.chromosome, interval = region.replace(',', '').split(':')
                try:
                    # see if just integer chromosomes are used
                    self.chromosome = int(self.chromosome)
                except ValueError:
                    pass
                self.begin, self.end = [int(i) for i in interval.split('-')]
                self.start = self.begin
            except ValueError:
                sys.exit("\n\nERROR: Window (-w/--window) inproperly formatted, "
                         "examples of accepted formats are:\n"
                         "'chr5:150200605-150423790' or 'ENST00000647408'\n\n")
            self.size = self.end - self.begin
            self.string = f"{self.chromosome}_{self.begin}_{self.end}"
        else:  # When region is an entire chromosome, contig or transcript
            if fasta is None:
                sys.exit("A fasta reference file is required if --window "
                         "is an entire chromosome, contig or transcript")
            else:
                from pyfaidx import Fasta
                self.chromosome = region
                self.begin = 0
                self.string = region
                self.end = len(Fasta(fasta)[region])
                self.size = self.end


def make_windows(full_window, max_size=1e6, fasta=None):
    reg = Region(full_window, fasta)
    if reg.size > max_size:
        chunks = ceil(reg.size / max_size)
        chsize = ceil(reg.size / chunks)
        return [
            Region(f"{reg.chromosome}:{reg.begin + i * chsize}-{reg.begin + (i + 1) * chsize}")
            for i in range(chunks)]
    else:
        return [reg]


def flatten(nested_list):
    return list(chain.from_iterable(nested_list))


def get_args():
    parser = ArgumentParser(description="plotting nanopolish methylation calls or frequency")
    parser.add_argument("-v", "--version",
                        help="Print version and exit.",
                        action="version",
                        version=f'methplotlib {__version__}')
    parser.add_argument("-m", "--methylation",
                        nargs='+',
                        help="data in nanopolish, nanocompore, ont-cram or bedgraph format",
                        required=True if "--example" not in sys.argv else False)
    parser.add_argument("-n", "--names",
                        nargs='+',
                        help="names of datasets in --methylation",
                        required=True if "--example" not in sys.argv else False)
    parser.add_argument("-w", "--window",
                        help="window (region) to which the visualisation has to be restricted",
                        required=True if "--example" not in sys.argv else False)
    parser.add_argument("-g", "--gtf",
                        help="add annotation based on a gtf file")
    parser.add_argument("-b", "--bed",
                        help="add annotation based on a bed file")
    parser.add_argument("-f", "--fasta",
                        help="required when --window is an entire chromosome, contig or transcript")
    parser.add_argument("--simplify",
                        help="simplify annotation track to show genes rather than transcripts",
                        action="store_true")
    parser.add_argument("--split",
                        help="split, rather than overlay the methylation tracks",
                        action="store_true")
    parser.add_argument("--static",
                        help="Make a static image of the browser window")
    parser.add_argument("--binary",
                        help="Make the nanopolish plot ignorning log likelihood nuances",
                        action="store_true")
    parser.add_argument("--smooth",
                        help="Rolling window size for averaging frequency values",
                        type=int,
                        default=5)
    parser.add_argument("--dotsize",
                        help="Control the size of dots in the per read plots",
                        type=int,
                        default=4)
    parser.add_argument("--minqual",
                        help="The minimal phred quality to show [for bam/cram input only]",
                        type=int,
                        default=20)
    parser.add_argument("--example",
                        action="store_true",
                        help="Show example command and exit.")
    parser.add_argument("-o", "--outfile",
                        help="File to write results to. "
                             "Default: methylation_browser_{chr}_{start}_{end}.html. "
                             "Use {region} as a shorthand for {chr}_{start}_{end} in the filename. "
                             "Missing paths will be created.")
    parser.add_argument("-q", "--qcfile",
                        help="File to write the qc report to. "
                             "Default: The path in outfile prefixed with qc_, "
                             "default is qc_report_methylation_browser_{chr}_{start}_{end}.html. "
                             "Use {region} as a shorthand for {chr}_{start}_{end} in the filename. "
                             "Missing paths will be created.")
    parser.add_argument("--store",
                        help=SUPPRESS,
                        action="store_true")
    args = parser.parse_args()
    if not args.example and not len(args.names) == len(args.methylation):
        sys.exit("INPUT ERROR: Expecting the same number of names as datasets!")
    return args


def init_logs(args):
    """Initiate log file and log arguments."""
    start_time = dt.fromtimestamp(time()).strftime('%Y%m%d_%H%M')
    logname = "methplotlib_" + start_time + ".log"
    handlers = [logging.FileHandler(logname)]
    logging.basicConfig(
        format='%(asctime)s %(message)s',
        handlers=handlers,
        level=logging.INFO)
    logging.info(f'methplotlib {__version__} started.')
    py_version = sys.version.replace('\n', ' ')
    logging.info(f'Python version is: {py_version}')
    logging.info(f'Arguments are: {args}')


def print_example():
    import pkg_resources
    meth = pkg_resources.resource_filename("methplotlib", "examples/ACTB_calls.tsv.gz")
    meth_freq = pkg_resources.resource_filename("methplotlib", "examples/meth_freq.tsv.gz")
    bed = pkg_resources.resource_filename("methplotlib", "examples/DNAse_cluster.bed.gz")
    annotation = pkg_resources.resource_filename("methplotlib", "examples/g38_locus.gtf.gz")

    example = f"""
methplotlib -m {meth} \\
               {meth_freq} \\
            -n calls frequencies \\
            -w chr7:5,525,542-5,543,028 \\
            -g {annotation} \\
            --simplify \\
            -b {bed} \\
            -o '{{region}}/example.html'""".strip()
    print(example)
    sys.exit(0)


def is_gz_file(filepath):
    import binascii
    with open(filepath, 'rb') as test_f:
        return binascii.hexlify(test_f.read(2)) == b'1f8b'


def is_cram_file(filepath):
    with open(filepath, 'rb') as test_f:
        return test_f.read(4) == b'CRAM'


def is_bam_file(filepath):
    import gzip
    try:
        with gzip.open(filepath) as test_f:
            return test_f.read(3) == b'BAM'
    except OSError:
        return False


def file_sniffer(filename):
    """
    Takes in a filename and tries to guess the input file type
    """
    if not Path(filename).is_file():
        sys.exit(f"\n\nERROR: File {filename} does not exist, please check the path!\n")
    if is_cram_file(filename):
        return "cram"
    if is_bam_file(filename):
        return "bam"
    if is_gz_file(filename):
        import gzip
        header = gzip.open(filename, 'rt').readline()
    else:
        header = open(filename, 'r').readline()

    if "GMM_anova_pvalue" in header:
        return "nanocompore"
    if "log_lik_methylated" in header:
        if "PS" in header:
            return "nanopolish_phased"
        else:
            return "nanopolish_call"
    if "num_motifs_in_group" in header:
        return "nanopolish_freq"
    if len(header.split('\t')) == 4:
        return 'bedgraph'  # there is no header, but the file is tab separated and has 4 fields
    sys.exit(f"\n\n\nInput file {filename} not recognized!\n")


def create_subplots(num_methrows, split, names=None, annotation=True):
    '''
    Prepare the panels (rows * 1 column) for the subplots.
    If splitting: one row for each dataset, taking 90%/len(datasets) for heights
    If not: one row spanning 4 rows and taking 90% of the heights
    if annotation is True (bed or gtf) then add a row with height 10%
    '''
    if split:
        return plotly.subplots.make_subplots(
            rows=num_methrows + annotation,
            cols=1,
            shared_xaxes=True,
            specs=[[{}] for i in range(num_methrows + annotation)],
            print_grid=False,
            subplot_titles=names,
            vertical_spacing=0.1 if num_methrows < 10 else 0.01,
            row_heights=[0.9 / num_methrows] * num_methrows + [0.1] * annotation

        )
    else:
        return plotly.subplots.make_subplots(
            rows=num_methrows + annotation,
            cols=1,
            shared_xaxes=True,
            specs=[[{'rowspan': num_methrows}], [None], [None], [None]] + [[{}]] * annotation,
            print_grid=False,
            vertical_spacing=0.1 if num_methrows < 10 else 0.01,
            row_heights=[0.9, 0, 0, 0] + [0.1] * annotation
        )


def create_browser_output(fig, outfile, window):
    if outfile is None:
        outfile = f"methylation_browser_{window.string}.html"
    else:
        from pathlib import Path

        outfile = outfile.format(region=window.string)
        p = Path(outfile)
        Path.mkdir(p.parent, exist_ok=True, parents=True)

    if outfile.endswith(".html"):
        write_html_output(fig, outfile)
    else:
        try:
            fig.write_image(outfile)
        except ValueError as e:
            sys.stderr.write("\n\nERROR: creating the image in this file format failed.\n")
            sys.stderr.write("ERROR: creating in default html format instead.\n")
            sys.stderr.write("ERROR: additional packages required. Detailed error:\n")
            sys.stderr.write(str(e))
            write_html_output(fig, outfile)


def write_html_output(fig, outfile):
    with open(outfile, "w+") as output:
        output.write(plotly.offline.plot(fig,
                                         output_type="div",
                                         show_link=False,
                                         include_plotlyjs='cdn'))



from argparse import ArgumentParser
import bgzip
import pandas as pd


def main():
    args = get_args()

    if args.naive:
        phased_calls = bgzip.open(args.phased_methylation, 'rt')

        phase1 = bgzip.open(args.prefix + "_phase1.tsv.gz", 'wt')
        phase2 = bgzip.open(args.prefix + "_phase2.tsv.gz", 'wt')
        uphase = bgzip.open(args.prefix + "_unphased.tsv.gz", 'wt')

        header = next(phased_calls)
        for output in [phase1, phase2, uphase]:
            output.write(header)

        for line in phased_calls:
            if line.endswith('1.0\n') or line.endswith('1\n'):
                phase1.write(line)
            elif line.endswith('2.0\n') or line.endswith('2\n'):
                phase2.write(line)
            else:
                uphase.write(line)
    else:
        df = pd.read_csv(args.phased_methylation, sep="\t")
        df[df["HP"].isna()].to_csv(args.prefix + "_calls_unphased.tsv.gz", sep="\t", index=False)
        df = df[df["HP"].notna()]
        ps_counts = df.loc[:, ["PS", "HP"]].drop_duplicates()["PS"].value_counts()
        hom_blocks = ps_counts[ps_counts == 1].index
        df[df.loc[:, "PS"].isin(hom_blocks)] \
            .to_csv(args.prefix + "_calls_homozygous.tsv.gz", sep="\t", index=False)
        df = df[~df.loc[:, "PS"].isin(hom_blocks)]
        df[df["HP"] == 1.0].to_csv(args.prefix + "_calls_phase1.tsv.gz", sep="\t", index=False)
        df[df["HP"] == 2.0].to_csv(args.prefix + "_calls_phase2.tsv.gz", sep="\t", index=False)


def get_args():
    parser = ArgumentParser(description="Split file with phased calls by phase.")
    parser.add_argument("phased_methylation", help="File created by annotate_calls_by_phase.py")
    parser.add_argument("-p", "--prefix", help="Prefix for output files", required=True)
    parser.add_argument("--naive",
                        help="Naively split reads, not taking homozygous regions into account.")
    return parser.parse_args()


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# Part of nanopolish
# Copied here for convenience
# Slightly edited by wdecoster


import sys
import csv
import argparse
import gzip


class SiteStats:
    def __init__(self, g_size, g_seq):
        self.num_reads = 0
        self.called_sites = 0
        self.called_sites_methylated = 0
        self.group_size = g_size
        self.sequence = g_seq


def update_call_stats(key, num_called_cpg_sites, is_methylated, sequence):
    if key not in sites:
        sites[key] = SiteStats(num_called_cpg_sites, sequence)

    sites[key].num_reads += 1
    sites[key].called_sites += num_called_cpg_sites
    if is_methylated > 0:
        sites[key].called_sites_methylated += num_called_cpg_sites


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Calculate methylation frequency per site')
    parser.add_argument('-c', '--call-threshold', type=float, required=False, default=2)
    parser.add_argument('-i', '--input', type=str, required=False)
    parser.add_argument('-s', '--split-groups', action='store_true')
    args = parser.parse_args()
    assert(args.call_threshold is not None)

    sites = dict()

    if args.input:
        if args.input.endswith(".gz"):
            in_fh = gzip.open(args.input, 'rt')
        else:
            in_fh = open(args.input)
    else:
        in_fh = sys.stdin

    for record in csv.DictReader(in_fh, delimiter='\t'):

        num_sites = int(record['num_motifs'])
        llr = float(record['log_lik_ratio'])

        # Skip ambiguous call
        if abs(llr) < args.call_threshold * num_sites:
            continue
        sequence = record['sequence']

        is_methylated = llr > 0

        # if this is a multi-cpg group and split_groups is set, break up these sites
        if args.split_groups and num_sites > 1:
            c = str(record['chromosome'])
            s = int(record['start'])
            e = int(record['end'])

            # find the position of the first CG dinucleotide
            sequence = record['sequence']
            cg_pos = sequence.find("CG")
            first_cg_pos = cg_pos
            while cg_pos != -1:
                key = (c, s + cg_pos - first_cg_pos, s + cg_pos - first_cg_pos)
                update_call_stats(key, 1, is_methylated, "split-group")
                cg_pos = sequence.find("CG", cg_pos + 1)
        else:
            key = (str(record['chromosome']), int(record['start']), int(record['end']))
            update_call_stats(key, num_sites, is_methylated, sequence)

    # header
    print("\t".join(["chromosome", "start", "end", "num_motifs_in_group", "called_sites",
                     "called_sites_methylated", "methylated_frequency", "group_sequence"]))

    sorted_keys = sorted(sites.keys(), key=lambda x: x)
    if len(sorted_keys) == 0:
        sys.exit("ERROR: No sites found for calculating frequencies!")

    for key in sorted_keys:
        if sites[key].called_sites > 0:
            (c, s, e) = key
            f = float(sites[key].called_sites_methylated) / sites[key].called_sites
            print("%s\t%s\t%s\t%d\t%d\t%d\t%.3f\t%s" % (
                c, s, e, sites[key].group_size, sites[key].called_sites, sites[key].called_sites_methylated, f, sites[key].sequence))



import pandas as pd
from pyranges import PyRanges


def methylation_pyranges_from_csv(inputfile):
    colnames = ["Chromosome", "Start", "End", "calls", "methylated"]
    return PyRanges(pd.read_csv(inputfile,
                                sep="\t",
                                names=colnames,
                                header=0,
                                usecols=[0, 1, 2, 4, 5]),
                    int64=True)



from methplotlib.import_methylation import get_data
from argparse import ArgumentParser
import sys
import methplotlib.qc as qc


def main():
    args = get_args()
    meth_data = get_data(args.methylation, args.names, None, 5)
    data = [m.table.rename({"methylated_frequency": m.name}, axis='columns')
            for m in meth_data if m.data_type == "frequency"]
    labels = [m.name for m in meth_data if m.data_type == "frequency"]
    full = data[0].join(data[1:]).dropna(how="any", axis="index")
    with open("qc_report_genome-wide.html", 'w') as qc_report:
        qc_report.write(qc.num_sites_bar(meth_data))
        qc_report.write(qc.pairwise_correlation_plot(full, labels))
        qc_report.write(qc.pca(full, labels))
        qc_report.write(qc.global_box(data))


def get_args():
    parser = ArgumentParser(description="genome wide splom")
    parser.add_argument("-m", "--methylation",
                        nargs='+',
                        help="nanopolish methylation calls or frequency output",
                        required=True)
    parser.add_argument("-n", "--names",
                        nargs='+',
                        help="names of datasets in --methylation",
                        required=True)
    args = parser.parse_args()
    if not len(args.names) == len(args.methylation):
        sys.exit("INPUT ERROR: Expecting the same number of names as datasets!")
    return args


if __name__ == '__main__':
    main()



import pandas as pd
from argparse import ArgumentParser
from statsmodels.sandbox.stats import multicomp


def main():
    args = get_args()
    df = pd.read_csv(args.test_result, sep="\t")
    df["padj"] = bhCorrection(df["p-value"])
    print(df.sort_values(by="padj").to_csv(sep="\t", index=False))


def bhCorrection(s):
    """
    Benjamini-Hochberg correction for a Series of p-values.
    """
    s = s.fillna(1.)
    q = multicomp.multipletests(s, method='fdr_bh')[1][:len(s)]
    q = pd.Series(q[:len(s)], s.index, name='p_adj')
    return q


def get_args():
    parser = ArgumentParser(
        description="Perform multiple testing correction and sort.")
    parser.add_argument("test_result",
                        help="File with results of differential modification "
                             "or allele-specific modification test.")
    return parser.parse_args()


if __name__ == '__main__':
    main()



import methplotlib.plots as plots
import methplotlib.utils as utils
import methplotlib.qc as qc
from methplotlib.import_methylation import get_data
import logging
import sys


def main():
    args = utils.get_args()
    if args.example:
        utils.print_example()
    utils.init_logs(args)
    windows = utils.make_windows(args.window, fasta=args.fasta)
    for window in windows:
        logging.info(f"Processing {window.string}")
        meth_data = get_data(args.methylation, args.names, window, args.smooth)
        if args.store:
            import pickle
            pickle.dump(
                obj=meth_data,
                file=open(f"methplotlib-data-{window.string}.pickle", 'wb'))
        logging.info(f"Collected methylation data for {len(meth_data)} datasets")
        qc.qc_plots(meth_data, window, qcpath=args.qcfile, outpath=args.outfile)
        logging.info("Created QC plots")
        meth_browser(meth_data=meth_data,
                     window=window,
                     gtf=args.gtf,
                     bed=args.bed,
                     simplify=args.simplify,
                     split=args.split,
                     outfile=args.outfile,
                     dotsize=args.dotsize,
                     static=args.static,
                     binary=args.binary,
                     minqual=args.minqual,  # only for input in bam/cram format
                     )
    logging.info("Finished!")


def meth_browser(meth_data, window, gtf=False, bed=False, simplify=False,
                 split=False, outfile=None, dotsize=4, static=False, binary=False, minqual=20):
    """
    meth_Data is a list of Methylation objects from the import_methylation submodule
    annotation is optional and is a gtf or bed file

    if the traces are to be --split per sample (or include raw data as flagged in data.split),
     then show one line per sample and one for the annotation, with methrows = number of datasets
    if no splitting is needed,
     then 4/5 of the browser is used for overlayed samples and one for gtf annotation
    the trace to be used for annotation is thus always num_methrows + 1
    """
    meth_traces = plots.methylation(meth_data, dotsize=dotsize, binary=binary, minqual=minqual)
    logging.info("Prepared methylation traces.")
    if split or meth_traces.split:
        num_methrows = len(meth_data)
        logging.info(f'Making browser in split mode, with {num_methrows} modification rows.')
        annot_row = num_methrows + 1
        annot_axis = f'yaxis{annot_row}'
        fig = utils.create_subplots(num_methrows,
                                    split=True,
                                    names=meth_traces.names,
                                    annotation=bool(bed or gtf))
        for y, (sample_traces, sample_type) in enumerate(meth_traces, start=1):
            logging.info(f"Adding traces of type {sample_type} at height {y}")
            for meth_trace in sample_traces:
                fig.append_trace(trace=meth_trace, row=y, col=1)
            if sample_type == 'nanopolish_freq':
                fig["layout"][f"yaxis{y}"].update(title="Modified <br> frequency", range=[0, 1])
                fig["layout"].update(showlegend=False)
                fig["layout"].update(legend=dict(orientation='h'))
            elif sample_type in ['nanopolish_call', 'nanopolish_phased']:
                fig["layout"][f"yaxis{y}"].update(title="Reads", tickformat=',d')
                fig["layout"].update(showlegend=False)
            elif sample_type == 'nanocompore':
                fig["layout"][f"yaxis{y}"].update(title="-log10(pval)")
                fig["layout"].update(legend=dict(orientation='h'))
            elif sample_type == 'ont-cram':
                fig["layout"][f"yaxis{y}"].update(title="Reads", tickformat=',d')
            elif sample_type == 'bedgraph':
                fig["layout"][f"yaxis{y}"].update(title="Value")
            else:
                sys.exit(f"ERROR: unrecognized data type {sample_type}")
    else:
        logging.info('Making browser in overlaying mode.')
        num_methrows = 4
        annot_row = 5
        annot_axis = 'yaxis2'
        fig = utils.create_subplots(num_methrows, split=False, annotation=bool(bed or gtf))
        for meth_trace in meth_traces.traces:
            for trace in meth_trace:
                fig.append_trace(trace=trace, row=1, col=1)
        fig["layout"].update(legend=dict(orientation='h'))
        if meth_traces.types[0] == 'nanopolish_freq':
            fig["layout"]['yaxis'].update(title="Modified <br> frequency", range=[0, 1])
        elif meth_traces.types[0] == 'bedgraph':
            fig["layout"]["yaxis"].update(title="Value")
        else:
            sys.exit(f"ERROR: unexpectedly not splitting for input of type {sample_type}")
    logging.info("Prepared modification plots.")

    if bed:
        for annot_trace in plots.bed_annotation(bed, window):
            fig.append_trace(trace=annot_trace, row=annot_row, col=1)
        y_max = -2
    if gtf:
        annotation_traces, y_max = plots.gtf_annotation(gtf, window, simplify)
        for annot_trace in annotation_traces:
            fig.append_trace(trace=annot_trace, row=annot_row, col=1)
    if bed or gtf:
        fig["layout"][annot_axis].update(range=[-2, y_max + 1],
                                         showgrid=False,
                                         zeroline=False,
                                         showline=False,
                                         ticks='',
                                         showticklabels=False)
        logging.info("Prepared annotation plots.")
    fig["layout"]["xaxis"].update(tickformat='g',
                                  separatethousands=True,
                                  range=[window.begin, window.end])
    fig["layout"].update(barmode='overlay',
                         title="Nucleotide modifications",
                         hovermode='closest',
                         plot_bgcolor='rgba(0,0,0,0)')
    if num_methrows > 10:
        for i in fig['layout']['annotations']:
            i['font']['size'] = 10
    utils.create_browser_output(fig, outfile, window)
    if static:
        import plotly.io as pio
        pio.write_image(fig, static, engine="kaleido")


if __name__ == '__main__':
    main()



import pandas as pd
import pyranges as pr
import numpy as np
import sys
import logging
from methplotlib.utils import file_sniffer, flatten
from itertools import repeat


class Modification(object):
    def __init__(self, table, data_type, name, called_sites, start_end_table=None):
        self.table = table
        self.data_type = data_type
        self.name = name
        self.called_sites = called_sites
        self.start_end_table = start_end_table


def get_data(methylation_files, names, window, smoothen=5):
    """
    Import methylation data from all files in the list methylation_files

    Data can in various formats
    - nanopolish raw, phased, frequency
    - cram
    - bam
    - nanocompore
    - bedgraph

    data is extracted within the window args.window
    Frequencies are smoothened using a sliding window
    """
    return flatten([read_mods(f, n, window, smoothen) for f, n in zip(methylation_files, names)])


def read_mods(filename, name, window, smoothen=5):
    """
    converts a file from nanopolish to a pandas dataframe
    input can be from calculate_methylation_frequency
    which will return a dataframe with 'chromosome', 'pos', 'methylated_frequency'
    smoothening the result by a rolling average

    input can also be raw data per read, optionally phased
    which will return a dataframe with 'read', 'chromosome', 'pos', 'log_lik_ratio', 'strand'
    """
    file_type = file_sniffer(filename)
    logging.info(f"File {filename} is of type {file_type}")
    try:
        if file_type.startswith("nanopolish"):
            return parse_nanopolish(filename, file_type, name, window, smoothen=smoothen)
        elif file_type == "nanocompore":
            return [parse_nanocompore(filename, name, window)]
        elif file_type in ["cram", "bam"]:
            return parse_cram(filename, file_type, name, window)
        elif file_type == 'bedgraph':
            return [parse_bedgraph(filename, name, window)]
    except Exception as e:
        logging.error(f"Error processing {filename}.")
        logging.error(e, exc_info=True)
        sys.stderr.write(f"\n\n\nError processing {filename}!\n")
        sys.stderr.write("\n\n\nDetailed error:\n")
        raise


def parse_nanopolish(filename, file_type, name, window, smoothen=5):
    if window:
        from pathlib import Path
        if Path(filename + '.tbi').is_file():
            import subprocess
            import gzip
            logging.info(f"Reading {filename} using a tabix stream.")
            region = f"{window.chromosome}:{window.begin}-{window.end}"
            try:
                tabix_stream = subprocess.Popen(['tabix', filename, region],
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE)
            except FileNotFoundError as e:
                logging.error("Error when opening a tabix stream.")
                logging.error(e, exc_info=True)
                sys.stderr.write("\n\nERROR when opening a tabix stream.\n")
                sys.stderr.write("Is tabix installed and on the PATH?.")
                raise
            header = gzip.open(filename, 'rt').readline().rstrip().split('\t')
            table = pd.read_csv(tabix_stream.stdout, sep='\t', header=None, names=header)
        else:
            logging.info(f"Reading {filename} slowly by splitting the file in chunks.")
            sys.stderr.write(f"\nReading {filename} would be faster with bgzip and tabix.\n")
            if file_type in ['nanopolish_call', 'nanopolish_phased']:
                sys.stderr.write("Please index with 'tabix -S1 -s1 -b3 -e4'.\n")
            else:
                sys.stderr.write("Please index with 'tabix -S1 -s1 -b2 -e3'.\n")
            iter_csv = pd.read_csv(filename, sep="\t", iterator=True, chunksize=1e6)
            table = pd.concat([chunk[chunk['chromosome'] == window.chromosome]
                               for chunk in iter_csv])
    else:
        table = pd.read_csv(filename, sep="\t")
    gr = pr.PyRanges(table.rename(columns={"start": "Start", "chromosome": "Chromosome",
                                           "end": "End", "Strand": "strand"}))
    logging.info("Read the file in a dataframe.")

    if window:
        gr = gr[str(window.chromosome), window.begin:window.end]
    try:
        gr.pos = np.floor(gr.drop().df[["Start", "End"]].mean(axis=1))
    except KeyError:
        sys.stderr.write(f"\n\n\nProblem parsing nanopolish file {filename}!\n")
        sys.stderr.write("Could it be that there are no calls in your selected window "
                         f"{window.chromosome}:{window.begin}-{window.end}?\n")
        sys.stderr.write("\n\n\nDetailed error:\n")
        raise

    table = gr.df

    if file_type in ['nanopolish_call', 'nanopolish_phased']:

        table = table.drop(columns=['Start', 'End', 'log_lik_methylated',
                                    'log_lik_unmethylated', 'num_calling_strands',
                                    'num_motifs', 'sequence'])
        if 'motif' in table:
            return [Modification(table=sub_df,
                                 data_type=file_type,
                                 name=f"{name}_{mod}",
                                 called_sites=len(sub_df),
                                 )
                    for mod, sub_df in table.groupby('motif')]
        else:
            return [Modification(
                table=table.sort_values(['read_name', 'pos']),
                data_type=file_type,
                name=name,
                called_sites=len(table))]
    if file_type == "nanopolish_freq":
        called_sites = table.called_sites
        table = table.drop(columns=['Start', 'End', 'num_motifs_in_group',
                                    'called_sites', 'called_sites_methylated',
                                    'group_sequence'])
        return [Modification(
            table=table.sort_values('pos')
                       .groupby('pos')
                       .mean()
                       .rolling(window=smoothen, center=True)
                       .mean(),
            data_type=file_type,
            name=name,
            called_sites=called_sites.sum())]


def parse_nanocompore(filename, name, window):
    def nanocompore_columns_of_interest(column):
        if column in ['pos', 'ref_id']:
            return True
        elif column.endswith('pvalue_context_2') or column.endswith('pvalue'):
            return True
        else:
            return False
    table = pd.read_csv(filename, sep="\t", usecols=nanocompore_columns_of_interest)
    if window:
        table = table[table["ref_id"] == window.chromosome]
    return Modification(
        table=table.sort_values('pos')
                   .append({'pos': window.end}, ignore_index=True)
                   .drop(columns="ref_id")
                   .fillna(1.0),
        data_type='nanocompore',
        name=name,
        called_sites=len(table))


def parse_bedgraph(filename, name, window):
    if window:
        from pathlib import Path
        if Path(filename + '.tbi').is_file():
            import subprocess
            logging.info(f"Reading {filename} using a tabix stream.")
            region = f"{window.chromosome}:{window.begin}-{window.end}"
            try:
                tabix_stream = subprocess.Popen(['tabix', filename, region],
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE)
            except FileNotFoundError as e:
                logging.error("Error when opening a tabix stream.")
                logging.error(e, exc_info=True)
                sys.stderr.write("\n\nERROR when opening a tabix stream.\n")
                sys.stderr.write("Is tabix installed and on the PATH?.")
                raise
            table = pd.read_csv(tabix_stream.stdout, sep='\t', header=None,
                                names=['Chromosome', 'Start', 'End', 'Value'])
        else:
            logging.info(f"Reading {filename} slowly by splitting the file in chunks.")
            sys.stderr.write(
                f"\nReading {filename} would be faster with bgzip and 'tabix -p bed'.\n")
            iter_csv = pd.read_csv(filename, sep="\t", iterator=True, chunksize=1e6, header=None,
                                   names=['Chromosome', 'Start', 'End', 'Value'])
            table = pd.concat([chunk[chunk['Chromosome'] == window.chromosome]
                               for chunk in iter_csv])
    else:
        table = pd.read_csv(filename, sep="\t", header=None,
                            names=['Chromosome', 'Start', 'End', 'Value'])
    gr = pr.PyRanges(table)
    logging.info("Read the file in a dataframe.")
    if window:
        gr = gr[window.chromosome, window.begin:window.end]
        if len(gr.df) == 0:
            sys.exit(f"No records for {filename} in {window.string}!\n")
    return Modification(
        table=gr.df.sort_values('Start'),
        data_type="bedgraph",
        name=name,
        called_sites=len(table))


def parse_cram(filename, filetype, name, window):
    import pysam
    mode = 'rc' if filetype == 'cram' else 'rb'
    cram = pysam.AlignmentFile(filename, mode)
    data = []
    start_stops = []
    for read in cram.fetch(reference=str(window.chromosome), start=window.begin, end=window.end):
        if not read.is_supplementary and not read.is_secondary:
            start_stops.append((read.query_name, read.reference_start, read.reference_end))
            mod_positions = get_modified_reference_positions(read)
            if mod_positions:
                data.extend(mod_positions)
    df = pd.DataFrame(data, columns=['read_name', 'strand', 'pos', 'quality', 'mod']) \
        .astype(dtype={'mod': 'category', 'quality': 'float'}) \
        .sort_values(['read_name', 'pos'])

    return [Modification(table=sub_df,
                         data_type="ont-cram",
                         name=f"{name}_{mod}",
                         called_sites=len(sub_df),
                         start_end_table=pd.DataFrame(start_stops,
                                                      columns=['read_name', 'posmin', 'posmax'])
                         .set_index('read_name')
                         )
            for mod, sub_df in df.groupby('mod')]


def get_modified_reference_positions(read):
    mod_positions = []
    if read.has_tag('Mm') or read.has_tag('MM'):
        modtag = 'Mm' if read.has_tag('Mm') else 'MM'
        qualtag = 'Ml' if read.has_tag('Ml') else 'ML'
        offset = 0
        for context in read.get_tag(modtag).split(';'):
            if context:
                basemod = context.split(',', 1)[0]
                if '-' in basemod:
                    sys.exit("ERROR: modifications on negative strand currently unsupported.\n"
                             "Please contact me if this would be of interest for you.")
                base, mod = basemod.split('+')
                # code below does not work with for ambiguous N and will search for a literal N
                if base == 'N':
                    sys.exit("ERROR: modifications of N nucleotides currently unsupported.\n"
                             "Please contact me if this would be of interest for you.")
                # The positions are encoded by specifying the number of non-modified occurences
                # of that specific bases to skip in the read sequence
                deltas = [int(i) for i in context.split(',')[1:]]
                if not deltas:
                    continue
                # Make an array with all positions for which a modification is reported
                # In coordinates relative to occurences of the nucleotide in the read
                locations = np.cumsum(deltas) + np.arange(len(deltas))
                # Nake an array with all read positions for which that nucleotide exists in the read
                # Oddly, get_forward_sequence() returns the sequence as it came from the sequencer
                # This is the same direction as the Mm/Ml tags are specified in
                base_index = np.array(
                    [i for i, letter in enumerate(read.get_forward_sequence()) if letter == base]
                )
                # Based on both arrays, find those read coordinates that have a modification
                modified_bases = base_index[locations]
                # Convert the array of read coordinates to reference coordinates
                # full_length=True will make sure the list is of the same length as the read
                # by adding 'None' for softclipped nucleotides
                refpos = np.array(read.get_reference_positions(full_length=True))
                # read.get_reference_positions() returns incrementing coordinates
                # regardless of strand, so reversing the order for reverse_stranded reads
                if read.is_reverse:
                    refpos = refpos[::-1]
                # The likelihoods are in an array of length of all Mm/MM deltas,
                # and are not separated by context/modified nucleotide type
                if read.has_tag(qualtag):
                    likelihoods = read.get_tag(qualtag).tolist()[offset:offset+len(deltas)]
                else:
                    likelihoods = [255] * len(deltas)
                offset += len(deltas)
                mod_positions.extend(
                    zip(repeat(read.query_name),  # read_name
                        repeat('-' if read.is_reverse else '+'),  # strand
                        refpos[modified_bases],  # pos
                        likelihoods,  # quality
                        repeat(basemod)  # mod
                        ))
        return mod_positions
    else:
        return None


def errs_tab(n):
    """Generate list of error rates for qualities less than equal than n."""
    return [10**(q / -10) for q in range(n + 1)]


def phred_to_probability(quals, tab=errs_tab(128)):
    return [tab[ord(q) - 33] for q in quals]



from sklearn.decomposition import PCA
import plotly.express as px
import pandas as pd
import plotly
import plotly.graph_objs as go


def qc_plots(meth_data, window, qcpath=None, outpath=None):

    if qcpath is None and outpath is None:
        outfile = f"qc_report_{window.string}.html"
    elif qcpath is None:
        from pathlib import Path, PosixPath
        p = Path(outpath.format(region=window.string))
        Path.mkdir(p.parent, exist_ok=True, parents=True)
        outfile = str(p.parent / PosixPath("qc_" + p.stem + ".html"))
    else:
        from pathlib import Path
        p = Path(qcpath)
        Path.mkdir(p.parent, exist_ok=True, parents=True)
        outfile = qcpath

    with open(outfile, 'w') as qc_report:
        qc_report.write(num_sites_bar(meth_data))
        if len([m for m in meth_data if m.data_type == "nanopolish_freq"]) > 0:
            data = [m.table.rename({"methylated_frequency": m.name}, axis='columns')
                    for m in meth_data if m.data_type == "nanopolish_freq"]
            full = data[0].join(data[1:]).dropna(how="any", axis="index")
            qc_report.write(modified_fraction_histogram(full))
        if len([m for m in meth_data if m.data_type == "nanopolish_freq"]) > 2:
            qc_report.write(pairwise_correlation_plot(full))
            qc_report.write(pca(full))
            qc_report.write(global_box(data))
        if len([m for m in meth_data
                if m.data_type in ["nanopolish_call", "nanopolish_phased"]]) > 2:
            pass


def num_sites_bar(meth_data):
    trace = go.Bar(x=[m.name for m in meth_data],
                   y=[m.called_sites for m in meth_data])
    layout = dict(title="Number of called positions")
    return plotly.offline.plot(dict(data=[trace], layout=layout),
                               output_type="div",
                               show_link=False,
                               include_plotlyjs='cdn')


def pairwise_correlation_plot(full):
    trace = go.Splom(dimensions=[dict(label=l, values=full[l]) for l in full.columns],
                     marker=dict(size=4,
                                 line=dict(width=0.5,
                                           color='rgb(230,230,230)')),
                     diagonal=dict(visible=False)
                     )

    layout = go.Layout(
        title='Correlation of modification frequency',
        dragmode='select',
        width=1200,
        height=1200,
        autosize=False,
        hovermode=False,
        plot_bgcolor='rgba(240,240,240, 0.95)')

    for i in range(1, len(full.columns) + 1):
        layout[f"xaxis{i}"] = dict(showline=True, zeroline=False, gridcolor='#fff', ticklen=4)
        layout[f"yaxis{i}"] = dict(showline=True, zeroline=False, gridcolor='#fff', ticklen=4)

    return plotly.offline.plot(dict(data=[trace], layout=layout),
                               output_type="div",
                               show_link=False,
                               include_plotlyjs='cdn')


def pca(full):
    sklearn_pca = PCA(n_components=2)
    pca = sklearn_pca.fit_transform(full.transpose())
    data = [dict(type='scatter',
                 x=[pca[index, 0]],
                 y=[pca[index, 1]],
                 mode='markers',
                 name=name,
                 hoverinfo='name',
                 marker=dict(
                     size=12,
                     line=dict(
                         color='rgba(217, 217, 217, 0.14)',
                         width=0.5),
                     opacity=0.8))
            for index, name in enumerate(full.columns)]

    layout = dict(xaxis=dict(title='PC1', showline=False),
                  yaxis=dict(title='PC2', showline=False),
                  title="Principal Component Analysis"
                  )
    return plotly.offline.plot(dict(data=data, layout=layout),
                               output_type="div",
                               show_link=False,
                               include_plotlyjs='cdn')


def global_box(data):
    fig = px.box(pd.concat([d.reset_index(drop=True)
                            .rename({d.columns[0]: "freq"}, axis="columns")
                            .assign(dataset=d.columns[0]) for d in data], ignore_index=True),
                 x="dataset", y="freq", title="Global frequency of modification")
    return plotly.offline.plot(fig,
                               output_type="div",
                               show_link=False,
                               include_plotlyjs='cdn')


def modified_fraction_histogram(full):
    traces = [go.Histogram(x=full[dataset].dropna(),
                           histnorm='probability density',
                           xbins=dict(start=0, size=0.01, end=1),
                           name=dataset,
                           opacity=0.6
                           )
              for dataset in full.columns]
    layout = dict(barmode="overlay",
                  title="Histogram of modified fractions",
                  xaxis=dict(title="Modified fraction"),
                  yaxis=dict(title="Frequency"))
    return plotly.offline.plot(dict(data=traces,
                                    layout=layout),
                               output_type="div",
                               show_link=False,
                               include_plotlyjs='cdn')



import pysam
from argparse import ArgumentParser
import gzip


def main():
    args = get_args()
    h1_reads = set([read.query_name for read in pysam.AlignmentFile(args.bamH1)])
    h2_reads = set([read.query_name for read in pysam.AlignmentFile(args.bamH2)])
    outputs = {f: gzip.open(f"{args.prefix}_{f}_meth.tsv.gz", 'wt') for f in ["H1", "H2", "U"]}
    meths = gzip.open(args.methylation, 'rt')
    header = next(meths)
    for o in outputs.values():
        o.write(header)
    for line in meths:
        outputs[find_read_in_sets(line.split('\t')[4], h1_reads, h2_reads)].write(line)


def find_read_in_sets(string, set_h1, set_h2):
    if string in set_h1:
        return "H1"
    elif string in set_h2:
        return "H2"
    else:
        return "U"


def get_args():
    parser = ArgumentParser(description="Split a nanopolish call-methylation file by haplotypes")
    parser.add_argument("methylation", help="File created by nanopolish call-methylation")
    parser.add_argument("--bamH1", help="bam file created by whatshap haplotag", required=True)
    parser.add_argument("--bamH2", help="bam file created by whatshap haplotag", required=True)
    parser.add_argument("--prefix", help="prefix for the output files", required=True)
    return parser.parse_args()


if __name__ == '__main__':
    main()



import pysam
from argparse import ArgumentParser
import gzip
import sys


def main():
    args = get_args()
    current_chrom = ""
    chrom_seen = []
    meth = gzip.open(args.methylation, 'rt')
    header = next(meth).rstrip().split('\t')
    header.extend(['PS', 'HP'])
    print('\t'.join(header))
    for pos in meth:
        line = pos.rstrip().split('\t')
        if line[0] != current_chrom:
            if line[0] in chrom_seen:
                sys.stderr.write("WARNING: this script is for chromosome-sorted meth files only!")
            sys.stderr.write(f"Switching to {line[0]}\n")
            phased_reads = get_phase_status_dict(bam=args.bam, chrom=line[0])
            current_chrom = line[0]
            chrom_seen.append(line[0])
        line.extend(phased_reads.get(line[4], ["NaN", "NaN"]))
        print('\t'.join(line))


def get_phase_status_dict(bam, chrom):
    return {read.query_name: [str(read.get_tag('PS')), str(read.get_tag('HP'))]
            for read in pysam.AlignmentFile(bam).fetch(contig=chrom)
            if read.has_tag('PS')}


def get_args():
    parser = ArgumentParser(description="Split a nanopolish call-methylation file by haplotypes")
    parser.add_argument("methylation", help="File created by nanopolish call-methylation")
    parser.add_argument("bam", help="bam file created by whatshap haplotag or longshot")
    return parser.parse_args()


if __name__ == '__main__':
    main()




import pandas as pd
import pyranges as pr
import numpy as np


def _methylated_and_freq_to_zero(df):

    only_in_bed = (df.methylated == -1) & (df.methylated_b == -1)
    df = df[~only_in_bed]

    a_zero = df.methylated == -1
    df.loc[a_zero, ["methylated", "calls"]] = 0
    b_zero = df.methylated_b == -1
    df.loc[b_zero, ["methylated_b", "calls_b"]] = 0

    return df


def count_total_and_methylated(df):
    """For each region in bed, get sum of calls and methylated"""

    grpby = df.groupby("ID")
    return pd.DataFrame({"Chromosome": df.Chromosome.iloc[0],
                         "Start": grpby.Start.first(),
                         "End": grpby.End.first(),
                         "ID": grpby.ID.first(),
                         "calls": grpby.calls.sum(),
                         "methylated": grpby.methylated.sum()})


def merge_regions_with_bed(bed, a, b):

    a = bed.join(a, how="left").apply(count_total_and_methylated).drop(like="_b")
    b = bed.join(b, how="left").apply(count_total_and_methylated).drop(like="_b")

    m = a.join(b, how="outer")

    return m


def main(a, b, bed):
    """1. Find the regions in bed that overlaps either a or/and b.
2. Sum methylations over regions in a/b that overlap bed.
3. Do fisher_exact on the methylation frequencies."""

    bed.ID = np.arange(len(bed))

    if "Strand" in bed:
        bed = bed.drop("Strand")

    m = merge_regions_with_bed(bed, a, b)

    m = m.apply(_methylated_and_freq_to_zero).drop(like="(Start|End|ID)_b|ID")

    m1, c1, m2, c2 = m.methylated, m.calls, m.methylated_b, m.calls_b,
    fe = pr.stats.fisher_exact(c1 - m1, m1, c2 - m2, m2, pseudocount=0.01)
    fe.insert(fe.shape[1], "FDR", pr.stats.fdr(fe.P))

    m = m.insert(fe[['OR', 'P', 'FDR']])

    return m



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
exec(open('methplotlib/version.py').read())

setup(
    name='methplotlib',
    version=__version__,
    description='Plot methylation data obtained from nanopolish',
    long_description=open(path.join(here, "README.md")).read(),
    long_description_content_type="text/markdown",
    scripts=["scripts/differential_modification", "scripts/allele_specific_modification"],
    url='https://github.com/wdecoster/methplotlib',
    author='Wouter De Coster',
    author_email='decosterwouter@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='nanopore',
    packages=find_packages(),
    python_requires='>=3',
    install_requires=['plotly>=4.9.0',
                      'numpy>=1.16.5',
                      'pandas>=0.23.4',
                      'pyranges>=0.0.77',
                      'fisher>=0.1.9',
                      'sklearn',
                      'pyfaidx',
                      'biopython',
                      'pysam'],
    package_data={'methplotlib': []},
    package_dir={'methplotlib': 'methplotlib'},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'methplotlib=methplotlib.methplotlib:main',
        ],
    },
)



import plotly.graph_objs as go
from methplotlib.annotation import parse_annotation, parse_bed
import sys
from sklearn.preprocessing import MinMaxScaler
import pandas as pd
import numpy as np


legend_made = False


class DataTraces(object):
    def __init__(self, traces, types, names, split):
        self.traces = traces
        self.types = types
        self.names = names
        self.split = split
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index == len(self.traces):
            raise StopIteration
        else:
            self.index += 1
            return self.traces[self.index - 1], self.types[self.index - 1]


def gtf_annotation(gtf, window, simplify=False):
    """
    Return a plotly trace for the annotation
    with a line for the entire gene and triangles for exons,
    indicating direction of transcription
    """
    result = []
    annotation = parse_annotation(gtf, window, simplify)
    if annotation:
        for y_pos, transcript in enumerate(annotation):
            line = make_per_gene_annot_line_trace(transcript, window, y_pos)
            exons = [make_per_exon_arrow_trace(transcript, begin, end, y_pos)
                     for begin, end in transcript.exon_tuples
                     if window.begin < begin and window.end > end]
            result.extend([line, *exons])
        return result, y_pos
    else:
        return result, 0


def make_per_gene_annot_line_trace(transcript, window, y_pos):
    """Generate a line trace for the gene

    Trace can get limited by the window sizes
    """
    return go.Scatter(x=[max(transcript.begin, window.begin),
                         min(transcript.end, window.end)],
                      y=[y_pos, y_pos],
                      mode='lines',
                      line=dict(width=2, color=transcript.color),
                      name=transcript.transcript,
                      text=transcript.gene,
                      hoverinfo='text',
                      showlegend=False)


def make_per_exon_arrow_trace(transcript, begin, end, y_pos):
    """Generate a line+marker trace for the exon

    The shape is an arrow, as defined by the strand in transcript.marker
    """
    return go.Scatter(x=[begin, end],
                      y=[y_pos, y_pos],
                      mode='lines+markers',
                      line=dict(width=8, color=transcript.color),
                      name=transcript.transcript,
                      text=transcript.gene,
                      hoverinfo='text',
                      showlegend=False,
                      marker=dict(symbol=transcript.marker,
                                  size=8))


def bed_annotation(bed, window):
    return [go.Scatter(x=[begin, end],
                       y=[-2, -2],
                       mode='lines',
                       line=dict(width=16, color='grey'),
                       text=name,
                       hoverinfo='text',
                       showlegend=False)
            for (begin, end, name) in parse_bed(bed, window)]


def methylation(meth_data, dotsize=4, binary=False, minqual=20):
    """
    Plot methylation traces from various data types
    """
    traces = []
    types = []
    names = []
    split = False
    for meth in meth_data:
        if meth.data_type in ['nanopolish_call', 'nanopolish_phased']:
            traces.append(
                make_per_read_meth_traces_llr(table=meth.table,
                                              phased=meth.data_type == 'nanopolish_phased',
                                              dotsize=dotsize,
                                              binary=binary)
            )
            split = True
        elif meth.data_type == 'nanocompore':
            traces.append(plot_nanocompore(meth.table, dotsize=dotsize))
            split = True
        elif meth.data_type == 'nanopolish_freq':
            traces.append(
                [go.Scatter(x=meth.table.index, y=meth.table["methylated_frequency"],
                            mode='lines',
                            name=meth.name,
                            hoverinfo='name')])
        elif meth.data_type == 'ont-cram':
            traces.append(
                make_per_read_meth_traces_phred(table=meth.table,
                                                minmax_table=meth.start_end_table,
                                                dotsize=dotsize,
                                                minqual=minqual)
            )
            split = True
        elif meth.data_type == 'bedgraph':
            starts = meth.table["Start"]
            ends = meth.table["End"]
            vals = meth.table["Value"]
            traces.append(
                [go.Scatter(x=[val for pair in zip(starts, ends) for val in pair],
                            y=[val for pair in zip(vals, vals) for val in pair],
                            mode='lines',
                            name=meth.name,
                            hoverinfo='name')])
        else:
            sys.exit(f"ERROR: unrecognized data type {meth.data_type}")
        types.append(meth.data_type)
        names.append(meth.name)
    return DataTraces(traces=traces,
                      types=types,
                      names=names,
                      split=split)


def make_per_read_meth_traces_phred(table, minmax_table, max_cov=100, dotsize=4, minqual=20):
    """Make traces for each read"""
    df_heights = assign_y_height_per_read(minmax_table, max_coverage=max_cov)
    table = table.join(df_heights, on="read_name", how="outer")
    traces = []
    hidden = 0
    for read in table["read_name"].unique():
        strand = table.loc[table["read_name"] == read, "strand"].values[0]
        try:
            traces.append(
                make_per_read_line_trace(read_range=minmax_table.loc[read],
                                         y_pos=df_heights.loc[read, 'height'],
                                         strand=strand)
            )
        except KeyError:
            hidden += 1
            continue
    if hidden:
        sys.stderr.write(f"Warning: hiding {hidden} reads because coverage above {max_cov}x.\n")
    traces.append(
        make_per_position_phred_scatter(read_table=table[table['quality'] > minqual],
                                        dotsize=dotsize)
    )
    return traces


def make_per_read_meth_traces_llr(table, phased=False, max_cov=100, dotsize=4, binary=False):
    """Make traces for each read"""
    minmax_table = find_min_and_max_pos_per_read(table, phased=phased)
    df_heights = assign_y_height_per_read(minmax_table, phased=phased, max_coverage=max_cov)
    table = table.join(df_heights, on="read_name")
    if binary:
        table.loc[:, "llr_scaled"] = binarize_log_likelihood_ratio(table["log_lik_ratio"].copy())
    else:
        table.loc[:, "llr_scaled"] = rescale_log_likelihood_ratio(table["log_lik_ratio"].copy())
    traces = []
    hidden = 0
    for read in table["read_name"].unique():
        strand = table.loc[table["read_name"] == read, "strand"].values[0]
        if phased:
            phase = table.loc[table["read_name"] == read, "HP"].values[0]
        else:
            phase = None
        try:
            traces.append(
                make_per_read_line_trace(read_range=minmax_table.loc[read],
                                         y_pos=df_heights.loc[read, 'height'],
                                         strand=strand,
                                         phase=phase,
                                         size=dotsize)
            )
        except KeyError:
            hidden += 1
            continue
    if hidden:
        sys.stderr.write(f"Warning: hiding {hidden} reads because coverage above {max_cov}x.\n")
    traces.append(
        make_per_position_likelihood_scatter(read_table=table, dotsize=dotsize)
    )
    return traces


def find_min_and_max_pos_per_read(table, phased=False):
    """Return a table with for every read the minimum and maximum position"""
    mm_table = table.loc[:, ["read_name", "pos"]] \
        .groupby('read_name') \
        .min() \
        .join(table.loc[:, ["read_name", "pos"]]
              .groupby('read_name')
              .max(),
              lsuffix="min",
              rsuffix="max")
    if phased:
        return mm_table.join(table.loc[:, ["read_name", "HP"]]
                             .drop_duplicates(subset='read_name')
                             .set_index('read_name'))
    else:
        return mm_table


def assign_y_height_per_read(df, phased=False, max_coverage=1000):
    """Assign height of the read in the per read traces

    Gets a dataframe of read_name, posmin and posmax.
    Sorting by position, and optionally by phase block.
    Determines optimal height (y coordinate) for this read

    Returns a dictionary mapping read_name to y_coord
    """
    if phased:
        dfs = df.sort_values(by=['HP', 'posmin', 'posmax'],
                             ascending=[True, True, False])
    else:
        dfs = df.sort_values(by=['posmin', 'posmax'],
                             ascending=[True, False])
    heights = [[] for i in range(max_coverage)]
    y_pos = dict()
    for read in dfs.itertuples():
        for y, layer in enumerate(heights, start=1):
            if len(layer) == 0:
                layer.append(read.posmax)
                y_pos[read.Index] = y
                break
            if read.posmin > layer[-1]:
                layer.append(read.posmax)
                y_pos[read.Index] = y
                break
    return pd.DataFrame({'read': list(y_pos.keys()),
                         'height': list(y_pos.values())}) \
        .set_index('read')


def rescale_log_likelihood_ratio(llr):
    """
    Rescale log likelihood ratios

    positive ratios between 0 and 1
    negative ratios between -1 and 0
    """
    scaler = MinMaxScaler(feature_range=(0, 1))
    llr[llr > 0] = scaler.fit_transform(llr[llr > 0].values.reshape(-1, 1)).tolist()
    scaler = MinMaxScaler(feature_range=(-1, 0))
    llr[llr < 0] = scaler.fit_transform(llr[llr < 0].values.reshape(-1, 1)).tolist()
    return llr


def binarize_log_likelihood_ratio(llr, cutoff=2):
    '''
    Converts the log likelihood to sorta binary representation,
    with 0 being insufficient evidence (based on cutoff)
    and +1 being probably methylated
    and -1 being probably unmethylated
    '''
    llr[(-cutoff < llr) & (llr < cutoff)] = 0
    llr[llr > cutoff] = 1
    llr[llr < -cutoff] = -1
    return llr


def make_per_read_line_trace(read_range, y_pos, strand, phase=None, size=4):
    """Make a grey line trace for a single read,
    with black arrow symbols on the edges indicating strand"""
    symbol = "triangle-right" if strand == "+" else "triangle-left"
    if phase:
        if phase == 1:
            color = 'lightgreen'
        elif phase == 2:
            color = 'yellow'
        else:  # phase is np.nan
            color = 'black'
    else:
        color = 'black'
    return go.Scatter(x=[read_range['posmin'], read_range['posmax']],
                      y=[y_pos, y_pos],
                      mode='lines+markers',
                      line=dict(width=1, color='lightgrey'),
                      showlegend=False,
                      marker=dict(symbol=symbol,
                                  size=size*2,
                                  color=color,
                                  line=dict(width=0.5,
                                            color='black')))


def make_per_position_likelihood_scatter(read_table, maxval=0.75, dotsize=4):
    """Make scatter plot per CpG per read
    with the RdBu colorscale from plotly 3.0.0, showing the
    scaled log likelihood of having methylation here"""
    old_RdBu = [[0, 'rgb(5,10,172)'],
                [0.35, 'rgb(106,137,247)'],
                [0.5, 'rgb(190,190,190)'],
                [0.6, 'rgb(220,170,132)'],
                [0.7, 'rgb(230,145,90)'],
                [1, 'rgb(178,10,28)']]
    return go.Scatter(x=read_table['pos'],
                      y=read_table['height'],
                      mode='markers',
                      showlegend=False,
                      text=read_table['log_lik_ratio'],
                      hoverinfo="text",
                      marker=dict(size=dotsize,
                                  color=read_table['llr_scaled'],
                                  cmin=-maxval,
                                  cmax=maxval,
                                  colorscale=old_RdBu,
                                  showscale=True,
                                  colorbar=dict(title="Modification likelihood",
                                                titleside="right",
                                                tickvals=[-maxval, 0, maxval],
                                                ticktext=["Likely <br> unmodified",
                                                          "0", "Likely <br> modified"],
                                                ticks="outside"))
                      )


def make_per_position_phred_scatter(read_table, dotsize=4):
    """Make scatter plot per CpG per read"""
    global legend_made
    if legend_made or read_table.dropna().empty:
        return go.Scatter(x=read_table['pos'],
                          y=read_table['height'],
                          mode='markers',
                          showlegend=False,
                          text=read_table['mod'],
                          hoverinfo="text",
                          marker=dict(size=dotsize,
                                      color=read_table['quality'],
                                      colorscale='Reds'))
    else:
        legend_made = True
        return go.Scatter(x=read_table['pos'],
                          y=read_table['height'],
                          mode='markers',
                          showlegend=False,
                          text=read_table['mod'],
                          hoverinfo="text",
                          marker=dict(size=dotsize,
                                      color=read_table['quality'],
                                      colorscale='Reds',
                                      colorbar=dict(title="Modification probability",
                                                    titleside="right",
                                                    tickvals=[read_table['quality'].min(),
                                                              read_table['quality'].max()],
                                                    ticktext=["Likely <br> unmodified",
                                                              "Likely <br> modified"],
                                                    ticks="outside")
                                      ))


def plot_nanocompore(table, dotsize=4):
    return [go.Scatter(x=table['pos'],
                       y=-table[pval].apply(np.log10),
                       mode='markers+lines',
                       name=pval,
                       marker=dict(size=dotsize))
            for pval in table.columns if not pval == 'pos']
